#  Copyright 2020-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union, List, Dict, Sequence, MutableSequence

from range_typed_integers import *

from skytemple_rust.st_sir0 import Sir0Serializable

_MappaFloorStructureType = u8
_MappaFloorWeather = u8
_MappaFloorDarknessLevel = u8
_MappaTrapType = u8
_MappaItemCategory = int
_MappaItem = int


class MappaTrapList:
    weights: Dict[_MappaTrapType, u16]

    def __init__(self, weights: Union[List[u16], Dict[_MappaTrapType, u16]]): ...

    def __eq__(self, other: object) -> bool: ...


class MappaMonster:
    level: u8
    main_spawn_weight: u16
    monster_house_spawn_weight: u16
    md_index: u16

    def __init__(self, level: u8, main_spawn_weight: u16, monster_house_spawn_weight: u16, md_index: u16):
        self.level = level
        self.main_spawn_weight = main_spawn_weight
        self.monster_house_spawn_weight = monster_house_spawn_weight
        self.md_index = md_index

    def __eq__(self, other: object) -> bool: ...


class MappaItemList:
    categories: Dict[_MappaItemCategory, int]
    items: Dict[_MappaItem, int]

    def __init__(
            self,
            categories: Dict[
                _MappaItemCategory, int
            ],
            items: Dict[_MappaItem, int],
    ):
        ...

    @classmethod
    def from_bytes(cls, data: bytes, pointer: int) -> MappaItemList: ...

    def to_bytes(self) -> bytes: ...

    def __eq__(self, other: object) -> bool: ...


class MappaFloorTerrainSettings:
    has_secondary_terrain: bool
    unk1: bool
    generate_imperfect_rooms: bool
    unk3: bool
    unk4: bool
    unk5: bool
    unk6: bool
    unk7: bool

    def __init__(
            self,
            has_secondary_terrain: bool,
            unk1: bool,
            generate_imperfect_rooms: bool,
            unk3: bool,
            unk4: bool,
            unk5: bool,
            unk6: bool,
            unk7: bool,
    ):
        ...

    def __eq__(self, other: object) -> bool: ...


class MappaFloorLayout:
    structure: _MappaFloorStructureType
    room_density: i8
    tileset_id: u8
    music_id: u8
    weather: _MappaFloorWeather
    floor_connectivity: u8
    initial_enemy_density: i8
    kecleon_shop_chance: u8
    monster_house_chance: u8
    unused_chance: u8
    sticky_item_chance: u8
    dead_ends: bool
    secondary_terrain: u8
    terrain_settings: MappaFloorTerrainSettings
    unk_e: bool
    item_density: u8
    trap_density: u8
    floor_number: u8
    fixed_floor_id: u8
    extra_hallway_density: u8
    buried_item_density: u8
    water_density: u8
    darkness_level: _MappaFloorDarknessLevel
    max_coin_amount: int
    kecleon_shop_item_positions: u8
    empty_monster_house_chance: u8
    unk_hidden_stairs: u8
    hidden_stairs_spawn_chance: u8
    enemy_iq: u16
    iq_booster_boost: i16

    def __init__(
            self,
            *,
            structure: _MappaFloorStructureType,
            room_density: i8,
            tileset_id: u8,
            music_id: u8,
            weather: _MappaFloorWeather,
            floor_connectivity: u8,
            initial_enemy_density: i8,
            kecleon_shop_chance: u8,
            monster_house_chance: u8,
            unusued_chance: u8,
            sticky_item_chance: u8,
            dead_ends: bool,
            secondary_terrain: u8,
            terrain_settings: MappaFloorTerrainSettings,
            unk_e: bool,
            item_density: u8,
            trap_density: u8,
            floor_number: u8,
            fixed_floor_id: u8,
            extra_hallway_density: u8,
            buried_item_density: u8,
            water_density: u8,
            darkness_level: _MappaFloorDarknessLevel,
            max_coin_amount: int,
            kecleon_shop_item_positions: u8,
            empty_monster_house_chance: u8,
            unk_hidden_stairs: u8,
            hidden_stairs_spawn_chance: u8,
            enemy_iq: u16,
            iq_booster_boost: i16,
    ):
        ...

    def __eq__(self, other: object) -> bool: ...


class MappaFloor:
    layout: MappaFloorLayout
    monsters: MutableSequence[MappaMonster]
    traps: MappaTrapList
    floor_items: MappaItemList
    shop_items: MappaItemList
    monster_house_items: MappaItemList
    buried_items: MappaItemList
    unk_items1: MappaItemList
    unk_items2: MappaItemList

    def __init__(
            self,
            layout: MappaFloorLayout,
            monsters: List[MappaMonster],
            traps: MappaTrapList,
            floor_items: MappaItemList,
            shop_items: MappaItemList,
            monster_house_items: MappaItemList,
            buried_items: MappaItemList,
            unk_items1: MappaItemList,
            unk_items2: MappaItemList,
    ):
        ...

    def __eq__(self, other: object) -> bool: ...


class MappaBin(Sir0Serializable):
    floor_lists: Sequence[Sequence[MappaFloor]]

    def __init__(self, floor_lists: List[List[MappaFloor]]): ...

    def add_floor_list(self, floor_list: List[MappaFloor]):...

    def remove_floor_list(self, index: int):...

    def add_floor_to_floor_list(self, floor_list_index: int, floor: MappaFloor): ...

    def insert_floor_in_floor_list(self, floor_list_index: int, insert_index: int, floor: MappaFloor): ...

    def remove_floor_from_floor_list(self, floor_list_index: int, floor_index: int):  ...

    def __eq__(self, other: object) -> bool: ...

class MappaBinWriter:
    def __new__(cls): ...
    def write(self, model: MappaBin) -> bytes: ...



#  Copyright 2020-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import Sequence, List, Tuple, Iterator

from range_typed_integers import *

_EvolutionMethod = u16
_AdditionalRequirement = u16
_Gender = u8
_PokeType = u8
_MovementType = u8
_IQGroup = u8
_Ability = u8
_ShadowSize = i8


class MdPropertiesState:
    num_entities: int
    max_possible: int

    @classmethod
    def instance(cls) -> "MdPropertiesState": ...


class MdEntry:
    md_index: u32
    entid: u16
    unk31: u16
    national_pokedex_number: u16
    base_movement_speed: u16
    pre_evo_index: u16
    evo_method: _EvolutionMethod
    evo_param1: u16
    evo_param2: _AdditionalRequirement
    sprite_index: i16
    gender: _Gender
    body_size: u8
    type_primary: _PokeType
    type_secondary: _PokeType
    movement_type: _MovementType
    iq_group: _IQGroup
    ability_primary: _Ability
    ability_secondary: _Ability
    exp_yield: u16
    recruit_rate1: i16
    base_hp: u16
    recruit_rate2: i16
    base_atk: u8
    base_sp_atk: u8
    base_def: u8
    base_sp_def: u8
    weight: i16
    size: i16
    unk17: u8
    unk18: u8
    shadow_size: _ShadowSize
    chance_spawn_asleep: i8
    hp_regeneration: u8
    unk21_h: i8
    base_form_index: i16
    exclusive_item1: i16
    exclusive_item2: i16
    exclusive_item3: i16
    exclusive_item4: i16
    unk27: i16
    unk28: i16
    unk29: i16
    unk30: i16
    bitfield1_0: bool
    bitfield1_1: bool
    bitfield1_2: bool
    bitfield1_3: bool
    can_move: bool
    bitfield1_5: bool
    can_evolve: bool
    item_required_for_spawning: bool

    @classmethod
    def new_empty(cls, entid: u16) -> "MdEntry": ...

    @property
    def md_index_base(self) -> int: ...


class Md:
    entries: Sequence[MdEntry]

    def __init__(self, data: bytes): ...

    def get_by_index(self, index: int) -> MdEntry: ...

    def get_by_entity_id(self, index: int) -> List[Tuple[int, MdEntry]]: ...

    def __len__(self) -> int: ...

    def __getitem__(self, key: int) -> MdEntry: ...

    def __setitem__(self, key: int, value: MdEntry) -> None: ...

    def __delitem__(self, key: int) -> None: ...

    def __iter__(self) -> Iterator[MdEntry]: ...

class MdWriter:
    def __new__(cls): ...
    def write(self, model: Md) -> bytes: ...



#  Copyright 2020-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import MutableSequence, Sequence

from range_typed_integers import *

from skytemple_rust.st_sir0 import Sir0Serializable


class LevelUpMove:
    move_id: u16
    level_id: u16

    def __init__(self, move_id: u16, level_id: u16):
        ...

    def __eq__(self, other: object) -> bool:
        ...


class MoveLearnset:
    level_up_moves: MutableSequence[LevelUpMove]
    tm_hm_moves: MutableSequence[u32]
    egg_moves: MutableSequence[u32]

    def __init__(self, level_up_moves: Sequence[LevelUpMove], tm_hm_moves: Sequence[u32], egg_moves: Sequence[u32]):
        ...

    def __eq__(self, other: object) -> bool:
        ...


class WazaMoveRangeSettings:
    target: int
    range: int
    condition: int
    unused: int

    def __init__(self, data: bytes):
        ...

    def __int__(self):
        ...

    def __eq__(self, other: object) -> bool:
        ...

class WazaMove:
    base_power: u16
    type: u8
    category: u8
    settings_range: WazaMoveRangeSettings
    settings_range_ai: WazaMoveRangeSettings
    base_pp: u8
    ai_weight: u8
    miss_accuracy: u8
    accuracy: u8
    ai_condition1_chance: u8
    number_chained_hits: u8
    max_upgrade_level: u8
    crit_chance: u8
    affected_by_magic_coat: bool
    is_snatchable: bool
    uses_mouth: bool
    ai_frozen_check: bool
    ignores_taunted: bool
    range_check_text: u8
    move_id: u16
    message_id: u8

    def __init__(self, data: bytes):
        ...

    def to_bytes(self) -> bytes:
        ...

    def __eq__(self, other: object) -> bool:
        ...


class WazaP(Sir0Serializable):
    moves: MutableSequence[WazaMove]
    learnsets: MutableSequence[MoveLearnset]

    def __init__(self, data: bytes, waza_content_pointer: int):
        ...

    def __eq__(self, other: object) -> bool:
        ...


class WazaPWriter:
    def __new__(cls): ...
    def write(self, model: WazaP) -> bytes: ...



#  Copyright 2021-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
class At3px:
    def __init__(self, data: bytes): ...
    def decompress(self) -> bytes: ...
    def to_bytes(self) -> bytes: ...
    @classmethod
    def cont_size(cls, data: bytes, byte_offset=0): ...
    @classmethod
    def compress(cls, data: bytes) -> 'At3px': ...



from __future__ import annotations

from typing import runtime_checkable, Protocol, Tuple, List, Optional, Any

from range_typed_integers import u32


@runtime_checkable
# See skytemple_files.container.sir0.sir0_serializable
class Sir0Serializable(Protocol):
    def sir0_serialize_parts(self) -> Tuple[bytes, List[u32], Optional[u32]]:
        ...

    @classmethod
    def sir0_unwrap(cls, content_data: bytes, data_pointer: u32) -> 'Sir0Serializable':
        ...


class Sir0:
    data_pointer: u32
    content: bytes
    content_pointer_offsets: List[u32]

    def __init__(
            self, content: bytes, pointer_offsets: List[u32], data_pointer: Optional[int] = None
    ):
        ...

    @classmethod
    def from_bin(cls, data: bytes) -> Sir0:
        ...


class Sir0Writer:
    def __new__(cls): ...
    def write(self, model: Sir0) -> bytes: ...




#  Copyright 2021-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Sequence, Optional

from PIL import Image

from skytemple_rust import TilemapEntry
from skytemple_rust.st_bpa import Bpa
from range_typed_integers import *


class BpcLayer:
    # The actual number of tiles is one lower
    number_tiles: u16
    # There must be 4 BPAs. (0 for not used)
    bpas: Sequence[u16]
    # NOTE: Incosistent with number_tiles. We are including the null chunk in this count.
    chunk_tilemap_len: u16
    # May also be set from outside after creation:
    tiles:  Sequence[bytes]
    tilemap: Sequence[TilemapEntry]

    def __init__(self, number_tiles: int, bpas: List[int], chunk_tilemap_len: int, tiles: List[bytes], tilemap: List[TilemapEntry]) -> None: ...


class Bpc:
    tiling_width: int
    tiling_height: int
    number_of_layers: int
    layers: Sequence[BpcLayer]

    def __init__(self, data: bytes, tiling_width: int, tiling_height: int): ...
    def chunks_to_pil(self, layer: int, palettes: Sequence[Sequence[int]], width_in_mtiles: int = 20) -> Image.Image: ...
    def single_chunk_to_pil(self, layer: int, chunk_idx: int, palettes: Sequence[Sequence[int]]) -> Image.Image: ...
    def tiles_to_pil(self, layer: int, palettes: Sequence[Sequence[int]], width_in_tiles: int = 20, single_palette: Optional[int] = None) -> Image.Image: ...

    def chunks_animated_to_pil(
            self, layer: int, palettes: Sequence[Sequence[int]], bpas: Sequence[Optional[Bpa]], width_in_mtiles: int = 20
    ) -> List[Image.Image]: ...

    def single_chunk_animated_to_pil(
            self, layer: int, chunk_idx: int, palettes: Sequence[Sequence[int]], bpas: Sequence[Optional[Bpa]]
    ) -> List[Image.Image]: ...

    def pil_to_tiles(self, layer: int, image: Image.Image) -> None: ...
    def pil_to_chunks(self, layer: int, image: Image.Image, force_import: bool = True) -> List[List[int]]: ...
    def get_tile(self, layer: int, index: int) -> TilemapEntry: ...
    def set_tile(self, layer: int, index: int, tile_mapping) -> None: ...  # : List[TilemapEntry]
    def get_chunk(self, layer: int, index: int) -> Sequence[TilemapEntry]: ...
    def import_tiles(self, layer: int, tiles: List[bytes], contains_null_tile: bool = False) -> None: ...

    def import_tile_mappings(
            self, layer: int, tile_mappings,  # : List[TilemapEntry]
            contains_null_chunk: bool = False, correct_tile_ids: bool = True
    ) -> None: ...

    def get_bpas_for_layer(self, layer: int, bpas_from_bg_list: Sequence[Optional[Bpa]]) -> List[Bpa]: ...
    def set_chunk(self, layer: int, index: int, new_tilemappings) -> None: ...  # : List[TilemapEntry]
    def remove_upper_layer(self) -> None: ...
    def add_upper_layer(self) -> None: ...
    def process_bpa_change(self, bpa_index: int, tiles_bpa_new: u16) -> None: ...


class BpcWriter:
    def __new__(cls): ...
    def write(self, model: Bpc) -> bytes: ...



#  Copyright 2020-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import Sequence, Union, Optional


class SmdlHeader:
    version: int
    unk1: int
    unk2: int
    modified_date: bytes
    file_name: bytes
    unk5: int
    unk6: int
    unk8: int
    unk9: int


class SmdlSong:
    unk1: int
    unk2: int
    unk3: int
    unk4: int
    tpqn: int
    unk5: int
    nbchans: int
    unk6: int
    unk7: int
    unk8: int
    unk9: int
    unk10: int
    unk11: int
    unk12: int


class SmdlEoc:
    param1: int
    param2: int


class SmdlTrackHeader:
    param1: int
    param2: int


class SmdlTrackPreamble:
    track_id: int
    channel_id: int
    unk1: int
    unk2: int


class SmdlEventPlayNote:
    velocity: int
    octave_mod: int
    note: int
    key_down_duration: Optional[int]


class SmdlEventPause:
    value: int


class SmdlEventSpecial:
    op: int
    params: Sequence[int]


class SmdlTrack:
    header: SmdlTrackHeader
    preamble: SmdlTrackPreamble
    events: Sequence[Union[SmdlEventSpecial, SmdlEventPause, SmdlEventPlayNote]]


class Smdl:
    header: SmdlHeader
    song: SmdlSong
    tracks: Sequence[SmdlTrack]
    eoc: SmdlEoc

    def __init__(self, data: bytes): ...


class SmdlWriter:
    def __new__(cls): ...
    def write(self, model: Smdl) -> bytes: ...



#  Copyright 2020-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Sequence


class SwdlPcmd:
    chunk_data: bytes


class SwdlPcmdReference:
    pcmd: SwdlPcmd
    offset: int
    length: int


class SwdlSampleInfoTblEntry:
    id: int
    ftune: int
    ctune: int
    rootkey: int  # seems unused by game!
    ktps: int
    volume: int  # (0-127)
    pan: int  # (0-64-127)
    unk5: int  # probably key_group, always 0
    unk58: int
    sample_format: int  # compare against SampleFormatConsts
    unk9: int
    loops: bool
    unk10: int
    unk11: int
    unk12: int
    unk13: int
    sample_rate: int
    sample: Optional[SwdlPcmdReference]
    loop_begin_pos: int  # (For ADPCM samples, the 4 bytes preamble is counted in the loopbeg!)
    loop_length: int

    envelope: int
    envelope_multiplier: int
    unk19: int
    unk20: int
    unk21: int
    unk22: int
    attack_volume: int
    attack: int
    decay: int
    sustain: int
    hold: int
    decay2: int
    release: int
    unk57: int


class SwdlWavi:
    sample_info_table: Sequence[Optional[SwdlSampleInfoTblEntry]]


class SwdlLfoEntry:
    unk34: int
    unk52: int
    dest: int
    wshape: int
    rate: int
    unk29: int
    depth: int
    delay: int
    unk32: int
    unk33: int


class SwdlSplitEntry:
    id: int
    unk11: int
    unk25: int
    lowkey: int
    hikey: int
    lolevel: int
    hilevel: int
    unk16: int
    unk17: int
    sample_id: int
    ftune: int
    ctune: int
    rootkey: int
    ktps: int
    sample_volume: int
    sample_pan: int
    keygroup_id: int
    unk22: int
    unk23: int
    unk24: int

    envelope: int
    envelope_multiplier: int
    unk37: int
    unk38: int
    unk39: int
    unk40: int
    attack_volume: int
    attack: int
    decay: int
    sustain: int
    hold: int
    decay2: int
    release: int
    unk53: int


class SwdlProgramTable:
    id: int
    prg_volume: int
    prg_pan: int
    unk3: int
    that_f_byte: int
    unk4: int
    unk5: int
    unk7: int
    unk8: int
    unk9: int
    lfos: Sequence[SwdlLfoEntry]
    splits: Sequence[SwdlSplitEntry]


class SwdlPrgi:
    program_table: Sequence[Optional[SwdlProgramTable]] = []


class SwdlKeygroup:
    id: int
    poly: int
    priority: int
    vclow: int
    vchigh: int
    unk50: int
    unk51: int


class SwdlKgrp:
    keygroups: Sequence[SwdlKeygroup]


class SwdlPcmdLen:
    reference: Optional[int]
    external: bool


class SwdlHeader:
    version: int
    unk1: int
    unk2: int
    modified_date: bytes
    file_name: bytes
    unk13: int
    pcmdlen: SwdlPcmdLen
    unk17: int


class Swdl:
    header: SwdlHeader
    wavi: SwdlWavi
    pcmd: Optional[SwdlPcmd]
    prgi: Optional[SwdlPrgi]
    kgrp: Optional[SwdlKgrp]

    def __init__(self, data: bytes): ...


class SwdlWriter:
    def __new__(cls): ...
    def write(self, model: Swdl) -> bytes: ...



from typing import Protocol, Dict, Optional, Sequence, List, TypeVar, Tuple, Any
from enum import Enum

exps_available = False
try:
    from explorerscript.source_map import SourceMap
    exps_available = True
except ImportError:
    pass


class HasIdAndNameProtocol(Protocol):
    id: int
    name: str


class ScriptDirectionProtocol(Protocol):
    ssb_id: int
    name: str


class ScriptOpCodeArgumentProtocol(Protocol):
    id: int
    type: str
    name: str


class ScriptOpCodeRepeatingArgumentGroupProtocol(Protocol):
    id: int
    arguments: Sequence[ScriptOpCodeArgumentProtocol]


class ScriptOpCodeProtocol(HasIdAndNameProtocol, Protocol):
    params: int
    stringidx: int
    unk2: int
    unk3: int
    arguments: Sequence[ScriptOpCodeArgumentProtocol]
    repeating_argument_group: Optional[ScriptOpCodeRepeatingArgumentGroupProtocol]


class ScriptDataProtocol(Protocol):
    @property
    def game_variables__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def game_variables__by_name(self) -> Dict[str, HasIdAndNameProtocol]: ...
    @property
    def objects__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def face_names__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def face_position_modes__by_id(self) -> List[HasIdAndNameProtocol]: ...
    @property
    def directions__by_ssb_id(self) -> Dict[int, ScriptDirectionProtocol]: ...
    @property
    def common_routine_info__by_id(self) -> List[HasIdAndNameProtocol]: ...
    @property
    def menus__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def process_specials__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def sprite_effects__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def bgms__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def level_list__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def level_entities__by_id(self) -> Dict[int, HasIdAndNameProtocol]: ...
    @property
    def op_codes__by_id(self) -> Dict[int, ScriptOpCodeProtocol]: ...


class SourceMapV2Protocol(Protocol):
    # todo
    pass


class SsbRoutineType(Enum):
    GENERIC = 1
    ACTOR = 3
    OBJECT = 4
    PERFORMER = 5
    COROUTINE = 9
    INVALID = -1


class SsbOperation(Protocol):
    offset: int
    op_code: HasIdAndNameProtocol
    params: List[Any]


class RoutineInfo(Protocol):
    type: SsbRoutineType
    linked_to: int
    linked_to_name: Optional[str] = None

    @property
    def linked_to_repr(self) -> Optional[str]: ...


class Ssb:
    original_binary_data: bytes
    routine_info: List[Tuple[int, RoutineInfo]]
    routine_ops: List[List[SsbOperation]]
    constants: List[str]
    strings: Dict[str, List[str]]

    @classmethod
    def create_empty(cls, scriptdata: ScriptDataProtocol, game_region: str) -> 'Ssb': ...
    def __init__(self, data: bytes, scriptdata: ScriptDataProtocol, game_region: str, string_codec: str): ...
    def to_explorerscript_v2(self) -> Tuple[str, SourceMapV2Protocol]: ...

    # NOTE: These two only work if exploerscript is installed. SsbOperation and RoutineInfo will be converted
    # to the appropriate explorerscript library types.
    if exps_available:
        def to_explorerscript(self) -> Tuple[str, 'SourceMap']: ...
        def to_ssb_script(self) -> Tuple[str, 'SourceMap']: ...


class SsbWriter:
    def __init__(self, script_data: ScriptDataProtocol, game_region: str, string_codec: str): ...
    def write(self, _model: Ssb) -> bytes: ...



#  Copyright 2021-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Sequence

from range_typed_integers import *


class BplAnimationSpec:
    duration_per_frame: u16
    number_of_frames: u16

    def __init__(self, duration_per_frame: u16, number_of_frames: u16): ...


class Bpl:
    number_palettes: u16
    has_palette_animation: bool
    palettes: Sequence[Sequence[int]]
    animation_specs: Sequence[BplAnimationSpec]
    animation_palette: Sequence[Sequence[int]]

    def __init__(self, data: bytes) -> None: ...
    def import_palettes(self, palettes: List[List[int]]) -> None: ...
    def apply_palette_animations(self, frame: int) -> List[List[int]]: ...
    def is_palette_affected_by_animation(self, pal_idx: int) -> bool: ...
    def get_real_palettes(self) -> List[List[int]]: ...
    def set_palettes(self, palettes: List[List[int]]) -> None: ...


class BplWriter:
    def __new__(cls): ...
    def write(self, model: Bpl) -> bytes: ...



#  Copyright 2021-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Tuple, Union, Iterator

from PIL.Image import Image


class KaoImage:
    @classmethod
    def create_from_raw(cls, cimg: bytes, pal: bytes) -> 'KaoImage': ...
    def get(self) -> Image: ...
    def clone(self) -> KaoImage: ...
    def size(self) -> int: ...
    def set(self, pil: Image) -> 'KaoImage': ...
    def raw(self) -> Tuple[bytes, bytes]: ...


class Kao:
    def __init__(self, data: bytes): ...
    @classmethod
    def create_new(cls, number_entries: int): ...
    def expand(self, new_size: int): ...
    def n_entries(self) -> int: ...
    def get(self, index: int, subindex: int) -> Optional[KaoImage]: ...
    def set(self, index: int, subindex: int, img: KaoImage): ...
    def set_from_img(self, index: int, subindex: int, pil: Image): ...
    def delete(self, index: int, subindex: int): ...
    def __iter__(self) -> Iterator[Tuple[int, int, Optional[KaoImage]]]: ...

class KaoWriter:
    def __new__(cls): ...
    def write(self, model: Kao) -> bytes: ...



#  Copyright 2020-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import Sequence, Tuple

from PIL import Image

from skytemple_rust import TilemapEntry
from skytemple_rust.st_dpci import Dpci


class Dpc:
    chunks: Sequence[Sequence[TilemapEntry]]

    def __init__(self, data: bytes): ...

    def chunks_to_pil(self, dpci: Dpci, palettes: Sequence[Sequence[int]], width_in_mtiles=16) -> Image.Image: ...

    def single_chunk_to_pil(self, chunk_idx, dpci: Dpci, palettes: Sequence[Sequence[int]]) -> Image.Image: ...

    def pil_to_chunks(self, image: Image.Image, force_import=True) -> Tuple[Sequence[bytes], Sequence[Sequence[int]]]: ...

    def import_tile_mappings(
            self, tile_mappings: Sequence[Sequence[TilemapEntry]],
            contains_null_chunk=False, correct_tile_ids=True
    ): ...


class DpcWriter:
    def __new__(cls): ...
    def write(self, model: Dpc) -> bytes: ...



#  Copyright 2021-2021 Parakoopa and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Tuple, Optional
from PIL.Image import Image

class WanImage:
    fragment_bytes_store: FragmentBytesStore
    frame_store: FrameStore
    animation_store: AnimationStore
    palette: Palette
    is_256_color: bool
    sprite_type: SpriteType
    unk2: int


class FragmentBytesStore:
    fragment_bytes: List[FragmentBytes]


class FragmentBytes:
    mixed_pixels: List[int]
    z_index: int

    def decode_fragment(self, resolution: FragmentResolution) -> List[int]: ...

    def to_image(self, palette: Palette, fragment: Fragment) -> List[int]: ...


class FrameStore:
    frames: List[Frame]
    max_fragment_alloc_count: int


class Fragment:
    unk1: int
    unk3_4: Optional[Tuple[bool, bool]]
    unk5: bool
    fragment_bytes_index: int
    offset_y: int
    offset_x: int
    flip: FragmentFlip
    is_mosaic: bool
    pal_idx: int
    resolution: FragmentResolution

class FragmentFlip:
    flip_h: bool
    flip_v: bool

class Frame:
    fragments: List[Fragment]
    frame_offset: Optional[FrameOffset]

class FrameOffset:
    head: Tuple[int, int]
    hand_left: Tuple[int, int]
    hand_right: Tuple[int, int]
    center: Tuple[int, int]

class FragmentResolution:
    x: int
    y: int


class AnimationStore:
    copied_on_previous: Optional[List[bool]]
    anim_groups: List[List[Animation]]


class Animation:
    frames: List[AnimationFrame]


class AnimationFrame:
    duration: int
    flag: int
    frame_id: int
    offset_x: int
    offset_y: int
    shadow_offset_x: int
    shadow_offset_y: int


class Palette:
    palette: List[int]


class SpriteType:
    PropsUI: SpriteType
    Chara: SpriteType
    Unknown: SpriteType
    name: str
    value: int

def encode_image_to_static_wan_file(image: Image) -> bytes: ...


#  Copyright 2021-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Optional, Sequence

from PIL import Image

from skytemple_rust.st_bpa import Bpa
from skytemple_rust.st_bpc import Bpc
from skytemple_rust.st_bpl import Bpl
from range_typed_integers import *


class Bma:
    map_width_camera: u8
    map_height_camera: u8
    tiling_width: u8
    tiling_height: u8
    map_width_chunks: u8
    map_height_chunks: u8
    number_of_layers: u16
    unk6: u16
    number_of_collision_layers: u16

    layer0: Sequence[int]
    layer1: Optional[Sequence[int]]

    # if unk6:
    unknown_data_block: Optional[Sequence[int]]
    # if number_of_collision_layers > 0:
    collision: Optional[Sequence[bool]]
    # if number_of_collision_layers > 1:
    collision2: Optional[Sequence[bool]]

    def __init__(self, data: bytes): ...

    def to_pil_single_layer(self, bpc: Bpc, palettes: Sequence[Sequence[int]], bpas: Sequence[Optional[Bpa]], layer: int) -> Image.Image: ...
    def to_pil(
            self, bpc: Bpc, bpl: Bpl, bpas: List[Optional[Bpa]],
            include_collision: bool = True, include_unknown_data_block: bool = True, pal_ani: bool = True, single_frame: bool = False
    ) -> List[Image.Image]: ...
    def from_pil(
            self, bpc: Bpc, bpl: Bpl, lower_img: Optional[Image.Image] = None, upper_img: Optional[Image.Image] = None,
            force_import: bool = False, how_many_palettes_lower_layer: int = 16
    ) -> None: ...
    def remove_upper_layer(self) -> None: ...
    def add_upper_layer(self) -> None: ...
    def resize(self, new_width_chunks: int, new_height_chunks: int, new_width_camera: int, new_height_camera: int) -> None: ...
    def place_chunk(self, layer_id: int, x: int, y: int, chunk_index: int) -> None: ...
    def place_collision(self, collision_layer_id: int, x: int, y: int, is_solid: bool) -> None: ...
    def place_data(self, x: int, y: int, data: int) -> None: ...
    def deepcopy(self) -> Bma: ...

class BmaWriter:
    def __new__(cls): ...
    def write(self, model: Bma) -> bytes: ...



from setuptools import setup
from setuptools_rust import Binding, RustExtension

setup(
    name="skytemple-rust",
    rust_extensions=[RustExtension(f"skytemple_rust", binding=Binding.PyO3)],
    packages=["skytemple_rust"],
    zip_safe=False,
)



#  Copyright 2020-2022 Capypara and the SkyTemple Contributors
#
#  This file is part of SkyTemple.
#
#  SkyTemple is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SkyTemple is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SkyTemple.  If not, see <https://www.gnu.org/licenses/>.
from typing import MutableSequence

from range_typed_integers import *

from skytemple_rust.st_sir0 import Sir0Serializable


class ItemPEntry:
    buy_price: u16
    sell_price: u16
    category: u8
    sprite: u8
    item_id: u16
    move_id: u16
    range_min: u8
    range_max: u8
    palette: u8
    action_name: u8
    is_valid: bool
    is_in_td: bool
    ai_flag_1: bool
    ai_flag_2: bool
    ai_flag_3: bool

    def __eq__(self, other: object) -> bool: ...


class ItemP(Sir0Serializable):
    item_list: MutableSequence[ItemPEntry]

    def __init__(self, data: bytes, pointer_to_pointers: int): ...

class ItemPWriter:
    def __new__(cls): ...
    def write(self, model: ItemP) -> bytes: ...
